package main

import (
	"bytes"
	"log"
	"path/filepath"
	"sort"
	"strconv"
	"time"

	"github.com/beevik/etree"
)

type Events []*etree.Element

func (e Events) Len() int      { return len(e) }
func (e Events) Swap(i, j int) { e[i], e[j] = e[j], e[i] }
func (e Events) Less(i, j int) bool {
	iDate, err := time.Parse(time.DateOnly, e[i].SelectAttrValue("eventDate", ""))
	if err != nil {
		log.Fatalln(err)
	}
	jDate, err := time.Parse(time.DateOnly, e[j].SelectAttrValue("eventDate", ""))
	if err != nil {
		log.Fatalln(err)
	}

	return iDate.Before(jDate)
}

type Merger struct {
	RegNumberDoc string
	Source       *etree.Element
	Data         map[string][]*etree.Element
}

func NewMerger() *Merger {
	return &Merger{
		Source: nil,
		Data:   make(map[string][]*etree.Element)}
}

func (m *Merger) EventCount() int {
	result := 0
	for _, events := range m.Data {
		result += len(events)
	}

	return result
}

func (m *Merger) AddFile(filePath string) error {
	log.Println("Чтение файла", filePath)

	doc := etree.NewDocument()
	if err := doc.ReadFromFile(filePath); err != nil {
		return err
	}

	if m.Source == nil {
		m.Source = doc.FindElement("/Document/Source")
	}

	if m.RegNumberDoc == "" {
		m.RegNumberDoc = doc.FindElement("/Document").SelectAttrValue("regNumberDoc", "")
	}

	for _, subject := range doc.FindElements("/Document/Data/*") {
		copySubject := subject.Copy()
		copySubject.RemoveChild(copySubject.FindElement("Events"))

		key := elementToString(copySubject)

		m.Data[key] = append(m.Data[key], subject.FindElements("Events/*")...)
	}

	return nil
}

func (m *Merger) WriteFile(filePath string) error {
	log.Println("Формирование результата")

	t := time.Now()

	// Сортируем события по дате
	for _, v := range m.Data {
		sort.Sort(Events(v))
	}

	doc := etree.NewDocument()

	doc.CreateProcInst("xml", `version="1.0" encoding="UTF-8"`)

	document := doc.CreateElement("Document")
	document.CreateAttr("dateDoc", t.Format(time.DateOnly))
	document.CreateAttr("groupBlocksCount", strconv.Itoa(m.EventCount()))
	document.CreateAttr("inn", INN)
	document.CreateAttr("ogrn", OGRN)
	document.CreateAttr("regNumberDoc", SOURCE_ID+t.Format("_20060102_150405"))
	document.CreateAttr("schemaVersion", SCHEMA_VERSION)
	document.CreateAttr("sourceID", SOURCE_ID)
	document.CreateAttr("subjectsCount", strconv.Itoa(len(m.Data)))

	document.AddChild(m.Source)

	data := document.CreateElement("Data")

	i := 0
	for k, v := range m.Data {
		subject := etree.NewDocument()
		err := subject.ReadFromString(k)
		if err != nil {
			return err
		}

		events := subject.FindElement("*").CreateElement("Events")

		for _, vv := range v {
			i++
			vv.CreateAttr("orderNum", strconv.Itoa(i))
			events.AddChild(vv)

		}
		data.AddChild(subject.FindElement("*"))
	}

	log.Println("Форматирование результата")
	doc.IndentTabs()
	doc.WriteSettings.CanonicalText = true

	log.Println("Запись результата в файл")
	return doc.WriteToFile(filePath)
}

func elementToString(e *etree.Element) string {
	b := new(bytes.Buffer)
	e.IndentWithSettings(&etree.IndentSettings{})
	e.WriteTo(b, &etree.WriteSettings{CanonicalText: true})

	return b.String()
}

func main() {
	merger := NewMerger()

	xmlFiles, err := filepath.Glob("*.xml")
	if err != nil {
		log.Fatalln(err)
	}

	for _, filePath := range xmlFiles {
		if filePath == "result.xml" {
			continue
		}

		err := merger.AddFile(filePath)
		if err != nil {
			log.Fatalln(err)
		}
	}

	merger.WriteFile("result.xml")

	log.Println("Завершено.")
}
